#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "lib.h"

#define FWD_OPCODE 0x0000000c

void usage(char *name)
{
  fprintf(stderr, "Usage: %s fw-address peer-address\n", name);
}

int main(int ac, char *av[])
{
  int s;
  unsigned char string_data[512];
  unsigned long long_data;
  unsigned long fw_addr, peer_addr;

  // debug_lib();

  if (ac != 3) {
    usage(av[0]);
    return 1;
  }

  if ((fw_addr = inet_addr(av[1])) == (unsigned long)-1) {
    fprintf(stderr, "Invalid firewall address specified\n");
    return 1;
  }

  if ((peer_addr = inet_addr(av[2])) == (unsigned long)-1) {
    fprintf(stderr, "Invalid peer address specified\n");
    return 1;
  }

  if ((s = part_one(fw_addr, peer_addr, FWD_OPCODE, "fwa1")) < 0)
    return 2;

  // Random number

  if (read_long(s, &long_data) < 0) {
    close(s);
    return 3;
  }

  if (long_data == 0) {
    fprintf(stderr, "Peer does not know us\n");
    close(s);
    return 3;
  }

  if (long_data > sizeof (string_data)) {
    fprintf(stderr, "Message is too long (%ld bytes)\n", long_data);
    return 3;
  }

  if (full_read(s, string_data, long_data) < 0) {
    close(s);
    return 3;
  }

  hex_dump("<- Authentication message", string_data, long_data);

  memset(find_hex(string_data, long_data, 32), 48, 32);

  if (write_long(s, long_data) < 0) {
    close(s);
    return 3;
  }

  if (full_write(s, string_data, long_data) < 0) {
    close(s);
    return 3;
  }

  hex_dump("-> Authentication message", string_data, long_data);

  if (read_long(s, &long_data) < 0) {
    close(s);
    return 3;
  }

  if (long_data == 0) {
    fprintf(stderr, "Authentication failed\n");
    close(s);
    return 3;
  }

  printf("<- OK (status = 0x%.8lx)\n", long_data);

  if (part_two(s) < 0)
    return 3;

  close(s);
  return 0;
}
