#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <sys/wait.h>

#include "global.h"
#include "md4.h"

#include "lib.h"

#define FWD_OPCODE 0x0000000c

void create_pw(int iter, char *addr, char *secret, char *buff)
{
  MD4_CTX ctx;
  unsigned char digest[16];
  int i, k;

  MD4Init(&ctx);
  MD4Update(&ctx, addr, strlen(addr));
  MD4Update(&ctx, secret, strlen(secret));
  MD4Final(digest, &ctx);

  for (i = 0; i < 8; i++)
    digest[i] = digest[i] ^ digest[i + 8];

  for (k = 0; k < iter; k++) {
    MD4Init(&ctx);
    MD4Update(&ctx, digest, 8);
    MD4Final(digest, &ctx);

    for (i = 0; i < 8; i++)
      digest[i] = digest[i] ^ digest[i + 8];
  }

  for (i = 0; i < 8; i++)
    sprintf(&buff[i << 1], "%02x", digest[i]);
}

void usage(char *name)
{
  fprintf(stderr, "Usage: %s fw-address peer-address secret\n", name);
}

int main(int ac, char *av[])
{
  int s;
  unsigned char string_data[256];
  unsigned long long_data;
  unsigned long fw_addr, peer_addr;

  // debug_lib();

  if (ac != 4) {
    usage(av[0]);
    return 1;
  }

  if ((fw_addr = inet_addr(av[1])) == (unsigned long)-1) {
    fprintf(stderr, "Invalid firewall address specified\n");
    return 1;
  }

  if ((peer_addr = inet_addr(av[2])) == (unsigned long)-1) {
    fprintf(stderr, "Invalid peer address specified\n");
    return 1;
  }

  if ((s = part_one(fw_addr, peer_addr, FWD_OPCODE, "skey")) < 0)
    return 2;

  if (read_long(s, &long_data) < 0) {
    close(s);
    return 3;
  }

  if (long_data == 0) {
    fprintf(stderr, "Peer does not know us (challenge = 0)\n");
    close(s);
    return 3;
  }

  printf("<- Challenge: %ld\n", long_data);

  create_pw(long_data, av[2], av[3], string_data);

  printf("-> Response: %s\n", string_data);

  long_data = strlen(string_data) + 1;

  if (write_long(s, long_data) < 0) {
    close(s);
    return 3;
  }

  if (full_write(s, string_data, long_data) < 0) {
    close(s);
    return 3;
  }

  if (write_long(s, 0) < 0) {
    close(s);
    return 3;
  }

  if (read_long(s, &long_data) < 0) {
    close(s);
    return 3;
  }

  if (long_data == 0) {
    fprintf(stderr, "Authentication failed\n");
    close(s);
    return 3;
  }

  printf("<- OK\n");

  if (part_two(s) < 0)
    return 3;

  close(s);
  return 0;
}

